"""
Hatch Plugin Registration
"""

from typing import Type

from hatchling.plugin import hookimpl

from hatch_pip_compile.plugin import PipCompileEnvironment


@hookimpl
def hatch_register_environment() -> Type[PipCompileEnvironment]:
    """
    Register the PipCompileEnvironment plugin with Hatch
    """
    return PipCompileEnvironment



"""
hatch-pip-compile header operations
"""

from __future__ import annotations

import hashlib
import logging
import pathlib
import re
from textwrap import dedent
from typing import Iterable

from packaging.requirements import Requirement
from packaging.version import Version
from piptools._compat.pip_compat import PipSession, parse_requirements

from hatch_pip_compile.base import HatchPipCompileBase
from hatch_pip_compile.exceptions import LockFileError

logger = logging.getLogger(__name__)


class PipCompileLock(HatchPipCompileBase):
    """
    Pip Compile Lock File Operations
    """

    def process_lock(self, lockfile: pathlib.Path) -> None:
        """
        Post process lockfile
        """
        version = f"{self.current_python_version.major}.{self.current_python_version.minor}"
        raw_prefix = f"""
        #
        # This file is autogenerated by hatch-pip-compile with Python {version}
        #
        """
        prefix = dedent(raw_prefix).strip()
        joined_dependencies = "\n".join([f"# - {dep}" for dep in self.environment.dependencies])
        lockfile_text = lockfile.read_text()
        cleaned_input_file = self.replace_temporary_lockfile(lockfile_text=lockfile_text)
        if self.environment.piptools_constraints_file is not None:
            lockfile_contents = self.environment.piptools_constraints_file.read_bytes()
            cross_platform_contents = lockfile_contents.replace(b"\r\n", b"\n")
            constraint_sha = hashlib.sha256(cross_platform_contents).hexdigest()
            constraints_path = self.environment.piptools_constraints_file.relative_to(
                self.environment.root
            ).as_posix()
            constraints_line = f"# [constraints] {constraints_path} (SHA256: {constraint_sha})"
            joined_dependencies = "\n".join([constraints_line, "#", joined_dependencies])
            cleaned_input_file = re.sub(
                r"-c \S*",
                lambda _: f"-c {constraints_path}",
                cleaned_input_file,
            )
        prefix += "\n" + joined_dependencies + "\n#"
        new_text = prefix + "\n\n" + cleaned_input_file
        lockfile.write_text(new_text)

    def read_header_requirements(self) -> list[Requirement]:
        """
        Read requirements from lock file header
        """
        lock_file_text = self.environment.piptools_lock_file.read_text()
        parsed_requirements = []
        for line in lock_file_text.splitlines():
            if line.startswith("# - "):
                requirement = Requirement(line[4:])
                parsed_requirements.append(requirement)
            elif not line.startswith("#"):
                break
        return parsed_requirements

    @property
    def current_python_version(self) -> Version:
        """
        Get python version

        In the case of running as a hatch plugin, the `virtualenv` will be set,
        otherwise it will be None and the Python version will be read differently.
        """
        if self.environment.virtual_env is not None:
            return Version(self.environment.virtual_env.environment["python_version"])
        else:
            msg = "VirtualEnv is not set"
            raise NotImplementedError(msg)

    @property
    def lock_file_version(self) -> Version:
        """
        Get lock file version
        """
        lock_file_text = self.environment.piptools_lock_file.read_text()
        match = re.search(
            r"# This file is autogenerated by hatch-pip-compile with Python (.*)", lock_file_text
        )
        if match is None:
            msg = "Could not find lock file python version"
            raise LockFileError(msg)
        return Version(match.group(1))

    def compare_python_versions(self, verbose: bool | None = None) -> bool:
        """
        Compare python versions

        Parameters
        ----------
        verbose : Optional[bool]
            Print warning if python versions are different, by default None
            which will print the warning. Used as a plugin flag.
        """
        lock_version = self.lock_file_version
        current_version = self.current_python_version
        match = (current_version.major == lock_version.major) and (
            current_version.minor == lock_version.minor
        )
        if match is False and verbose is not False:
            logger.error(
                "[hatch-pip-compile] Your Python version is different "
                "from the lock file, your results may vary."
            )
        return lock_version == current_version

    def compare_requirements(self, requirements: Iterable[Requirement]) -> bool:
        """
        Compare requirements

        Parameters
        ----------
        requirements : Iterable[Requirement]
            List of requirements to compare against the lock file
        """
        lock_requirements = self.read_header_requirements()
        return set(requirements) == set(lock_requirements)

    def compare_constraint_sha(self, sha: str) -> bool:
        """
        Compare SHA to the SHA on the lockfile
        """
        lock_file_text = self.environment.piptools_lock_file.read_text()
        match = re.search(r"# \[constraints\] \S* \(SHA256: (.*)\)", lock_file_text)
        if match is None:
            return False
        return match.group(1).strip() == sha.strip()

    def get_file_content_hash(self) -> str:
        """
        Get hash of lock file
        """
        lockfile_contents = self.environment.piptools_lock_file.read_bytes()
        cross_platform_contents = lockfile_contents.replace(b"\r\n", b"\n")
        return hashlib.sha256(cross_platform_contents).hexdigest()

    def read_lock_requirements(self) -> list[Requirement]:
        """
        Read all requirements from lock file
        """
        if not self.environment.dependencies:
            return []
        install_requirements = parse_requirements(
            str(self.environment.piptools_lock_file),
            session=PipSession(),
        )
        return [ireq.req for ireq in install_requirements]  # type: ignore[misc]

    def replace_temporary_lockfile(self, lockfile_text: str) -> str:
        """
        Replace the temporary lockfile with the new lockfile
        """
        cleaned_input_file = re.sub(
            rf"-r \S*[\\/]{self.environment.name}\.in",
            f"hatch.envs.{self.environment.name}",
            lockfile_text,
        )
        return cleaned_input_file



"""
Shared fixtures for tests.
"""

from __future__ import annotations

import contextlib
import os
import pathlib
import shutil
from dataclasses import dataclass, field
from subprocess import CompletedProcess
from typing import Generator
from unittest.mock import patch

import hatch
import pytest
import tomlkit
from click.testing import CliRunner, Result
from hatch.cli.application import Application
from hatch.config.constants import AppEnvVars, ConfigEnvVars, PublishEnvVars
from hatch.project.core import Project
from hatch.utils.fs import Path, temp_directory
from hatch.utils.platform import Platform

from hatch_pip_compile.plugin import PipCompileEnvironment


@pytest.fixture
def mock_check_command() -> Generator[patch, None, None]:
    """
    Disable the `plugin_check_command` for testing
    """
    with patch("hatch_pip_compile.plugin.PipCompileEnvironment.plugin_check_command") as mock:
        mock.return_value = CompletedProcess(args=[], returncode=0, stdout=b"", stderr=b"")
        yield mock


@pytest.fixture
def subprocess_run() -> Generator[patch, None, None]:
    """
    Disable the `subprocess.run` for testing
    """
    with patch("subprocess.run") as mock:
        mock.return_value = CompletedProcess(args=[], returncode=0, stdout=b"", stderr=b"")
        yield mock


@pytest.fixture
def platform() -> Platform:
    """
    Platform
    """
    return Platform()


@pytest.fixture
def isolation(platform: Platform) -> Generator[Path, None, None]:
    """
    Isolated hatch environment for testing.
    """
    with temp_directory() as temp_dir:
        data_dir = pathlib.Path(__file__).parent / "data"
        shutil.copytree(data_dir, temp_dir, dirs_exist_ok=True)
        data_dir = temp_dir / "data"
        data_dir.mkdir()
        cache_dir = temp_dir / "cache"
        cache_dir.mkdir()
        default_env_vars = {
            AppEnvVars.NO_COLOR: "1",
            ConfigEnvVars.DATA: str(data_dir),
            ConfigEnvVars.CACHE: str(cache_dir),
            PublishEnvVars.REPO: "dev",
            "HATCH_SELF_TESTING": "true",
            "PYAPP_COMMAND_NAME": os.urandom(4).hex(),
            "GIT_AUTHOR_NAME": "Foo Bar",
            "GIT_AUTHOR_EMAIL": "foo@bar.baz",
            "COLUMNS": "80",
            "LINES": "24",
        }
        if platform.windows:  # pragma: no cover
            default_env_vars["COMSPEC"] = "cmd.exe"
        else:
            default_env_vars["SHELL"] = "sh"

        with temp_dir.as_cwd(default_env_vars):
            os.environ.pop(AppEnvVars.ENV_ACTIVE, None)
            os.environ.pop(AppEnvVars.FORCE_COLOR, None)
            yield temp_dir


@dataclass
class PipCompileFixture:
    """
    Testing Fixture Data Container
    """

    __test__ = False

    isolation: pathlib.Path
    toml_doc: tomlkit.TOMLDocument
    pyproject: pathlib.Path
    project: Project
    platform: Platform
    isolated_data_dir: pathlib.Path

    application: Application = field(init=False)
    default_environment: PipCompileEnvironment = field(init=False)
    test_environment: PipCompileEnvironment = field(init=False)
    cli_runner: CliRunner = field(init=False)

    def __post_init__(self) -> None:
        """
        Post Init
        """
        self.application = Application(
            exit_func=lambda x: None,  # noqa: ARG005
            verbosity=0,
            interactive=False,
            enable_color=False,
        )
        self.application.data_dir = self.isolated_data_dir / "data"
        self.application.config_file.load()
        self.application.cache_dir = self.isolated_data_dir / "cache"
        self.application.project = self.project
        self.current_environment = self.reload_environment("default")
        self.default_environment = self.reload_environment("default")
        self.test_environment = self.reload_environment("test")
        self.cli_runner = CliRunner()

    def reload_environment(self, environment: str | PipCompileEnvironment) -> PipCompileEnvironment:
        """
        Reload a new environment given the current state of the isolated project
        """
        if isinstance(environment, PipCompileEnvironment):
            environment_name = environment.name
        else:
            environment_name = environment
        env = self.application.get_environment(env_name=environment_name)
        return env

    def update_pyproject(self) -> None:
        """
        Update pyproject.toml
        """
        tomlkit.dump(self.toml_doc, self.pyproject.open("w"))

    @contextlib.contextmanager
    def chdir(self) -> Generator[None, None, None]:
        """
        Change the working directory to the isolation
        """
        current_dir = os.getcwd()
        try:
            os.chdir(self.isolation)
            yield
        finally:
            os.chdir(current_dir)

    def update_environment_resolver(
        self, environment: str | PipCompileEnvironment, resolver: str
    ) -> PipCompileEnvironment:
        """
        Update the environment resolver
        """
        if isinstance(environment, PipCompileEnvironment):
            environment_name = environment.name
        else:
            environment_name = environment
        self.toml_doc["tool"]["hatch"]["envs"][environment_name]["pip-compile-resolver"] = resolver
        self.update_pyproject()
        return self.reload_environment(environment_name)

    def update_environment_installer(
        self, environment: str | PipCompileEnvironment, installer: str
    ) -> PipCompileEnvironment:
        """
        Update the environment installer
        """
        if isinstance(environment, PipCompileEnvironment):
            environment_name = environment.name
        else:
            environment_name = environment
        self.toml_doc["tool"]["hatch"]["envs"][environment_name][
            "pip-compile-installer"
        ] = installer
        self.update_pyproject()
        return self.reload_environment(environment_name)

    def cli_invoke(self, args: list[str], env: dict[str, str] | None = None) -> Result:
        """
        Invoke the CLI
        """
        invoke_kwargs = {"args": args}
        if env:
            invoke_kwargs["env"] = env
        with self.cli_runner.isolated_filesystem(self.isolation):
            return self.cli_runner.invoke(hatch.cli.hatch, **invoke_kwargs)

    def invoke_environment(
        self, environment: PipCompileEnvironment, env: dict[str, str] | None = None
    ) -> Result:
        """
        Sync the environment
        """
        result = self.cli_invoke(
            args=["env", "run", "--env", environment.name, "--", "python", "--version"],
            env=env,
        )
        return result

    @staticmethod
    def virtualenv_exists(environment: PipCompileEnvironment) -> bool:
        """
        Check if the virtual environment exists
        """
        virtualenv_cfg = environment.virtual_env.directory / "pyvenv.cfg"
        return virtualenv_cfg.exists()

    def is_installed(
        self,
        environment: PipCompileEnvironment,
        package: str,
    ) -> bool:
        """
        Check if a package is installed in the environment

        This method simply checks if the package is in the site-packages directory
        of the virtual environment.
        """
        if not self.virtualenv_exists(environment=environment):
            return False
        if self.platform.windows:
            site_packages = environment.virtual_env.directory / "Lib" / "site-packages"
            if not site_packages.exists():
                return False
        else:
            site_packages_search = environment.virtual_env.directory.glob(
                "lib/python*/site-packages"
            )
            site_packages = next(site_packages_search, None)
            if site_packages is None:
                return False
        package_dir = site_packages / package
        return (package_dir / "__init__.py").exists()


@pytest.fixture
def pip_compile(
    isolation: Path,
    platform: Platform,
) -> PipCompileFixture:
    """
    PipCompile testing fixture
    """
    pyproject = isolation / "pyproject.toml"
    isolated_data_dir = Path(os.environ[ConfigEnvVars.DATA])
    return PipCompileFixture(
        isolation=isolation,
        toml_doc=tomlkit.parse(string=pyproject.read_text()),
        pyproject=pyproject,
        project=Project(path=isolation),
        platform=platform,
        isolated_data_dir=isolated_data_dir,
    )


@pytest.fixture(autouse=True)
def pip_compile_disable(monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Delete the PIP_COMPILE_DISABLE environment variable
    """
    monkeypatch.delenv("PIP_COMPILE_DISABLE", raising=False)


resolver_param = pytest.mark.parametrize("resolver", ["pip-compile", "uv"])
installer_param = pytest.mark.parametrize("installer", ["pip", "pip-sync", "uv"])



"""
hatch-pip-compile CLI
"""

from __future__ import annotations

import dataclasses
import json
import os
import subprocess
from typing import Any, Sequence

import click
import rich.traceback

from hatch_pip_compile.__about__ import __application__, __version__


@dataclasses.dataclass
class HatchCommandRunner:
    """
    Hatch Command Runner
    """

    environments: Sequence[str] = dataclasses.field(default_factory=list)
    upgrade: bool = False
    upgrade_all: bool = False
    upgrade_packages: Sequence[str] = dataclasses.field(default_factory=list)

    former_env_vars: dict[str, Any] = dataclasses.field(init=False, default_factory=dict)
    console: rich.console.Console = dataclasses.field(init=False)
    supported_environments: set[str] = dataclasses.field(init=False)

    def __post_init__(self):
        """
        Initialize the internal state
        """
        self.console = rich.console.Console()
        rich.traceback.install(show_locals=True, console=self.console)
        self.supported_environments = self._get_supported_environments()
        if all(
            [not self.environments, "default" in self.supported_environments, not self.upgrade_all]
        ):
            self.environments = ["default"]
        elif not self.environments and not self.upgrade_all:
            msg = "Either `--all` or an environment name must be specified"
            raise click.BadParameter(msg)
        elif self.upgrade_all:
            self.environments = list(self.supported_environments)
        unsupported_environments = set(self.environments).difference(self.supported_environments)
        if unsupported_environments:
            msg = (
                f"The following environments are not supported or unknown: "
                f"{', '.join(unsupported_environments)}. "
                f"Supported environments are: {', '.join(sorted(self.supported_environments))}"
            )
            raise click.BadParameter(msg)

    def __enter__(self) -> HatchCommandRunner:
        """
        Set the environment variables
        """
        env_vars = {"__PIP_COMPILE_FORCE__": "1"}
        if self.upgrade:
            env_vars["PIP_COMPILE_UPGRADE"] = "1"
            self.console.print(
                "[bold green]hatch-pip-compile[/bold green]: Upgrading all dependencies"
            )
        elif self.upgrade_packages:
            env_vars["PIP_COMPILE_UPGRADE_PACKAGES"] = ",".join(self.upgrade_packages)
            message = (
                "[bold green]hatch-pip-compile[/bold green]: "
                f"Upgrading packages: {', '.join(self.upgrade_packages)}"
            )
            self.console.print(message)
        self.former_env_vars = {
            key: os.environ.get(key) for key in env_vars.keys() if os.environ.get(key) is not None
        }
        os.environ.update(env_vars)
        return self

    def __exit__(self, *args, **kwargs):
        """
        Restore the environment variables
        """
        os.environ.update(self.former_env_vars)

    def hatch_cli(self):
        """
        Run the `hatch` CLI
        """
        self.console.print(
            "[bold green]hatch-pip-compile[/bold green]: Targeting environments: "
            f"{', '.join(sorted(self.environments))}"
        )
        for environment in sorted(self.environments):
            environment_command = [
                "hatch",
                "env",
                "run",
                "--env",
                environment,
                "--",
                "python",
                "--version",
            ]
            self.console.print(
                f"[bold green]hatch-pip-compile[/bold green]: Running "
                f"`[bold blue]{' '.join(environment_command)}`[/bold blue]"
            )
            result = subprocess.run(
                args=environment_command,
                capture_output=True,
                check=False,
            )
            if result.returncode != 0:  # pragma: no cover
                self.console.print(
                    "[bold yellow]hatch command[/bold yellow]: "
                    f"[bold blue]`{' '.join(environment_command)}`[/bold blue]"
                )
                self.console.print(result.stdout.decode("utf-8"))
                self.console.print(
                    "[bold red]hatch-pip-compile[/bold red]: Error running hatch command"
                )
                raise click.exceptions.Exit(1)

    @classmethod
    def _get_supported_environments(cls) -> set[str]:
        """
        Get the names of the environments from `hatch env show --json`

        Returns
        -------
        List[str]
            The name of the environments
        """
        result = subprocess.run(
            args=["hatch", "env", "show", "--json"],
            capture_output=True,
            check=True,
        )
        environment_dict: dict[str, Any] = json.loads(result.stdout)
        return {
            key for key, value in environment_dict.items() if value.get("type") == "pip-compile"
        }


@click.command("hatch-pip-compile")
@click.version_option(version=__version__, prog_name=__application__)
@click.argument("environment", default=None, type=click.STRING, required=False, nargs=-1)
@click.option(
    "-U",
    "--upgrade/--no-upgrade",
    is_flag=True,
    default=False,
    help="Try to upgrade all dependencies to their latest versions",
)
@click.option(
    "-P",
    "--upgrade-package",
    "upgrade_packages",
    nargs=1,
    multiple=True,
    help="Specify a particular package to upgrade; may be used more than once",
)
@click.option(
    "--all",
    "upgrade_all",
    is_flag=True,
    default=False,
    help="Upgrade all environments",
)
def cli(
    environment: Sequence[str],
    upgrade: bool,
    upgrade_packages: Sequence[str],
    upgrade_all: bool,
):
    """
    Upgrade your `hatch-pip-compile` managed dependencies
    from the command line.
    """
    with HatchCommandRunner(
        environments=environment,
        upgrade=upgrade,
        upgrade_packages=upgrade_packages,
        upgrade_all=upgrade_all,
    ) as hatch_runner:
        hatch_runner.hatch_cli()


if __name__ == "__main__":
    cli()



"""
Installation Tests
"""

from unittest.mock import Mock

from tests.conftest import PipCompileFixture


def test_pip_install_dependencies(mock_check_command: Mock, pip_compile: PipCompileFixture) -> None:
    """
    Assert the `pip` installation command is called with the expected arguments
    """
    pip_compile.default_environment.create()
    pip_compile.default_environment.installer.install_dependencies()
    expected_call = [
        "python",
        "-u",
        "-m",
        "pip",
        "install",
        "--disable-pip-version-check",
        "--no-python-version-warning",
        "-q",
        "--requirement",
    ]
    call_args = list(mock_check_command.call_args)[0][0][:-1]
    assert call_args == expected_call



"""
Plugin tests.
"""

from unittest.mock import Mock

import pytest

from hatch_pip_compile.exceptions import HatchPipCompileError
from hatch_pip_compile.resolver import PipCompileResolver
from tests.conftest import PipCompileFixture


def test_lockfile_path(
    pip_compile: PipCompileFixture,
) -> None:
    """
    Test the default lockfile paths
    """
    assert (
        pip_compile.default_environment.piptools_lock_file
        == pip_compile.isolation / "requirements.txt"
    )
    assert (
        pip_compile.test_environment.piptools_lock_file
        == pip_compile.isolation / "requirements" / "requirements-test.txt"
    )


def test_piptools_constraints_file(
    pip_compile: PipCompileFixture,
) -> None:
    """
    Test constraints paths
    """
    assert pip_compile.default_environment.piptools_constraints_file is None
    assert (
        pip_compile.test_environment.piptools_constraints_file
        == pip_compile.default_environment.piptools_lock_file
    )


def test_expected_dependencies(pip_compile: PipCompileFixture) -> None:
    """
    Test expected dependencies from `PipCompileEnvironment`
    """
    assert set(pip_compile.default_environment.dependencies) == {"hatch"}
    assert set(pip_compile.test_environment.dependencies) == {"pytest", "pytest-cov", "hatch"}


def test_lockfile_up_to_date_missing(pip_compile: PipCompileFixture) -> None:
    """
    Test the `lockfile_up_to_date` property when the lockfile is missing
    """
    pip_compile.default_environment.piptools_lock_file.unlink()
    assert pip_compile.default_environment.lockfile_up_to_date is False
    assert pip_compile.default_environment.dependencies_in_sync() is False


def test_lockfile_up_to_date_empty(pip_compile: PipCompileFixture) -> None:
    """
    Test the `lockfile_up_to_date` property when the lockfile is empty
    """
    pip_compile.default_environment.piptools_lock_file.write_text("")
    assert pip_compile.default_environment.lockfile_up_to_date is False
    assert pip_compile.default_environment.dependencies_in_sync() is False


def test_lockfile_up_to_date_mismatch(pip_compile: PipCompileFixture) -> None:
    """
    Test the `lockfile_up_to_date` property when the lockfile is mismatched
    """
    lock_text = pip_compile.default_environment.piptools_lock_file.read_text()
    lock_text = lock_text.replace("# - hatch", "#")
    pip_compile.default_environment.piptools_lock_file.write_text(lock_text)
    assert pip_compile.default_environment.lockfile_up_to_date is False


def test_pip_compile_cli(mock_check_command: Mock, pip_compile: PipCompileFixture) -> None:
    """
    Test the `pip_compile_cli` method is called with the expected arguments
    """
    pip_compile.default_environment.pip_compile_cli()
    expected_call = [
        "python",
        "-m",
        "piptools",
        "compile",
        "--quiet",
        "--no-header",
        "--resolver=backtracking",
        "--strip-extras",
        "--output-file",
    ]
    call_args = list(mock_check_command.call_args)[0][0][:-2]
    assert call_args == expected_call


def test_env_var_disabled(pip_compile: PipCompileFixture, monkeypatch: pytest.MonkeyPatch) -> None:
    """
    Test the `lockfile_up_to_date` property when the lockfile is empty
    """
    monkeypatch.setenv("PIP_COMPILE_DISABLE", "1")
    with pytest.raises(HatchPipCompileError, match="attempted to run a lockfile update"):
        pip_compile.default_environment.pip_compile_cli()


@pytest.mark.parametrize("environment_name", ["default", "misc", "docs"])
def test_constraint_env_self(pip_compile: PipCompileFixture, environment_name: str) -> None:
    """
    Test the value of the constraint env b/w the default and test environments
    """
    environment = pip_compile.reload_environment(environment=environment_name)
    assert environment.constraint_env is environment


@pytest.mark.parametrize("environment_name", ["test"])
def test_constraint_env_other(pip_compile: PipCompileFixture, environment_name: str) -> None:
    """
    Test the value of the constraint env b/w the default and test environments
    """
    environment = pip_compile.reload_environment(environment=environment_name)
    assert environment.constraint_env.name == pip_compile.default_environment.name


@pytest.mark.parametrize("environment_name", ["default", "docs", "misc"])
def test_prepare_environment(pip_compile: PipCompileFixture, environment_name: str) -> None:
    """
    Test the `prepare_environment` method
    """
    environment = pip_compile.reload_environment(environment=environment_name)
    environment.prepare_environment()
    if environment.dependencies:
        assert environment.piptools_lock_file.exists()
    else:
        assert not environment.piptools_lock_file.exists()
    assert environment.dependencies_in_sync() is True
    assert environment.lockfile_up_to_date is True


def test_resolver_instance_default(pip_compile: PipCompileFixture) -> None:
    """
    Test the `pip-compile-resolver` option on default resolver
    """
    assert isinstance(pip_compile.default_environment.resolver, PipCompileResolver)


def test_resolver_instance_pip_compile(pip_compile: PipCompileFixture) -> None:
    """
    Test that the `PipCompileResolver` is used
    """
    pip_compile.toml_doc["tool"]["hatch"]["envs"]["default"]["pip-compile-resolver"] = "pip-compile"
    pip_compile.update_pyproject()
    environment = pip_compile.reload_environment("default")
    assert isinstance(environment.resolver, PipCompileResolver)



"""
Dependency Resolvers
"""

from __future__ import annotations

import os
from abc import ABC, abstractmethod
from typing import ClassVar

from hatch_pip_compile.base import HatchPipCompileBase


class BaseResolver(HatchPipCompileBase, ABC):
    """
    Base Resolver for the plugin
    """

    resolver_options: ClassVar[list[str]] = []

    @property
    @abstractmethod
    def resolver_executable(self) -> list[str]:
        """
        Resolver Executable
        """

    def get_pip_compile_args(self, input_file: os.PathLike, output_file: os.PathLike) -> list[str]:
        """
        Get the pip compile arguments
        """
        upgrade = bool(os.getenv("PIP_COMPILE_UPGRADE"))
        upgrade_packages = os.getenv("PIP_COMPILE_UPGRADE_PACKAGE") or None
        upgrade_args = []
        upgrade_package_args = []
        if upgrade:
            upgrade_args.append("--upgrade")
        if upgrade_packages:
            upgrade_packages_sep = upgrade_packages.split(",")
            for package in upgrade_packages_sep:
                upgrade_package_args.append(f"--upgrade-package={package.strip()}")
        cmd = [
            *self.resolver_executable,
            "--verbose"
            if self.environment.config.get("pip-compile-verbose", None) is True
            else "--quiet",
            "--no-header",
            *self.resolver_options,
        ]
        if self.environment.config.get("pip-compile-hashes", False) is True:
            cmd.append("--generate-hashes")
        if self.environment.piptools_constraints_file is not None:
            cmd.extend(["--constraint", str(self.environment.piptools_constraints_file)])
        cmd.extend(self.environment.config.get("pip-compile-args", []))
        cmd.extend(upgrade_args)
        cmd.extend(upgrade_package_args)
        cmd.extend(["--output-file", str(output_file), str(input_file)])
        return cmd


class PipCompileResolver(BaseResolver):
    """
    Pip Compile Resolver
    """

    pypi_dependencies: ClassVar[list[str]] = ["pip-tools"]
    resolver_options: ClassVar[list[str]] = ["--resolver=backtracking", "--strip-extras"]

    @property
    def resolver_executable(self) -> list[str]:
        """
        Resolver Executable
        """
        return [
            self.environment.virtual_env.python_info.executable,
            "-m",
            "piptools",
            "compile",
        ]


class UvResolver(BaseResolver):
    """
    Uv Resolver
    """

    pypi_dependencies: ClassVar[list[str]] = ["uv"]

    @property
    def resolver_executable(self) -> list[str]:
        """
        Resolver Executable
        """
        return [
            self.environment.virtual_env.python_info.executable,
            "-m",
            "uv",
            "pip",
            "compile",
        ]



"""
Integration Tests using the CLI
"""

import platform

import pytest

from hatch_pip_compile.exceptions import HatchPipCompileError
from tests.conftest import PipCompileFixture, installer_param, resolver_param


@installer_param
@resolver_param
@pytest.mark.parametrize("environment_name", ["default", "test", "lint", "docs", "misc"])
def test_invoke_environment_creates_env(
    pip_compile: PipCompileFixture, environment_name: str, resolver: str, installer: str
) -> None:
    """
    Test using the CLI runner
    """
    if installer == "pip-sync" and platform.system().lower() == "windows":
        pytest.skip("pip-sync + CLI runner not supported on Windows")
    environment = pip_compile.update_environment_resolver(
        environment=environment_name, resolver=resolver
    )
    environment = pip_compile.update_environment_installer(
        environment=environment, installer=installer
    )
    assert not pip_compile.virtualenv_exists(environment=environment)
    result = pip_compile.invoke_environment(environment=environment)
    assert result.exit_code == 0
    assert pip_compile.virtualenv_exists(environment=environment)


def test_missing_lockfile_created(pip_compile: PipCompileFixture) -> None:
    """
    Running the CLI without a lockfile creates one
    """
    environment = pip_compile.default_environment
    environment.piptools_lock_file.unlink()
    assert not environment.piptools_lock_file.exists()
    result = pip_compile.invoke_environment(environment=environment)
    assert result.exit_code == 0
    assert environment.piptools_lock_file.exists()
    assert pip_compile.virtualenv_exists(environment=environment)


def test_constraint_env_created(pip_compile: PipCompileFixture) -> None:
    """
    Running the CLI with a constraint env creates one
    """
    environment = pip_compile.test_environment
    environment.piptools_lock_file.unlink()
    environment.constraint_env.piptools_lock_file.unlink()
    result = pip_compile.invoke_environment(environment=environment)
    assert result.exit_code == 0
    assert environment.piptools_lock_file.exists()
    assert environment.constraint_env.piptools_lock_file.exists()
    assert environment.virtual_env.directory.exists()
    assert environment.constraint_env.virtual_env.directory.exists()


def test_missing_lockfile_after_prepared(pip_compile: PipCompileFixture) -> None:
    """
    After an environment is prepared the lockfile is deleted and recreated the next time
    """
    environment = pip_compile.default_environment
    # Create the environment the first time
    result = pip_compile.invoke_environment(environment=environment)
    assert result.exit_code == 0
    # Delete the lockfile
    assert environment.piptools_lock_file.exists()
    environment.piptools_lock_file.unlink()
    assert not environment.piptools_lock_file.exists()
    # Run the environment again
    result = pip_compile.invoke_environment(
        environment=environment,
    )
    assert result.exit_code == 0
    # Assert the lockfile was recreated
    assert environment.piptools_lock_file.exists()


@resolver_param
@pytest.mark.parametrize("environment_name", ["default", "test", "lint"])
def test_pip_compile_disable_cli(
    pip_compile: PipCompileFixture, environment_name: str, resolver: str
) -> None:
    """
    Test that the `PIP_COMPILE_DISABLE` environment variable raises an error
    """
    environment = pip_compile.update_environment_resolver(
        environment=environment_name, resolver=resolver
    )
    environment.piptools_lock_file.unlink(missing_ok=True)
    result = pip_compile.invoke_environment(
        environment=environment,
        env={"PIP_COMPILE_DISABLE": "1"},
    )
    assert result.exit_code == 1
    assert isinstance(result.exception, HatchPipCompileError)


def test_prune_removes_all_environments(pip_compile: PipCompileFixture) -> None:
    """
    Assert that running `hatch env prune` removes all environments
    """
    pip_compile.default_environment.create()
    pip_compile.test_environment.create()
    assert pip_compile.default_environment.virtual_env.directory.exists()
    assert pip_compile.test_environment.virtual_env.directory.exists()
    result = pip_compile.cli_invoke(args=["env", "prune"])
    assert result.exit_code == 0
    assert not pip_compile.default_environment.virtual_env.directory.exists()
    assert not pip_compile.test_environment.virtual_env.directory.exists()


@installer_param
@resolver_param
def test_add_new_dependency(pip_compile: PipCompileFixture, installer: str, resolver: str) -> None:
    """
    Create a new environment, assert that it exists without the `requests` package,
    then add the `requests` package to the environment, assert that it is installed.
    """
    if platform.system().lower() == "windows" and installer == "pip-sync":
        pytest.skip("pip-sync + CLI runner not supported on Windows")
    environment = pip_compile.default_environment
    assert not environment.virtual_env.directory.exists()
    pip_compile.update_environment_resolver(environment=environment, resolver=resolver)
    pip_compile.update_environment_installer(environment=environment, installer=installer)
    result = pip_compile.invoke_environment(environment=environment)
    assert result.exit_code == 0
    if result.exit_code != 0:
        raise result.exception
    assert environment.virtual_env.directory.exists()
    assert not pip_compile.is_installed(environment=environment, package="requests")
    assert "requests" not in environment.piptools_lock_file.read_text()
    pip_compile.toml_doc["project"]["dependencies"] += ["requests"]
    pip_compile.update_pyproject()
    new_result = pip_compile.invoke_environment(environment=environment)
    if new_result.exit_code != 0:
        raise new_result.exception
    assert "requests" in environment.piptools_lock_file.read_text()
    assert pip_compile.is_installed(environment=environment, package="requests")


@installer_param
@resolver_param
def test_add_new_dependency_constraint_env(
    pip_compile: PipCompileFixture, installer: str, resolver: str
) -> None:
    """
    Test dependency workflow in a constraint environment
    """
    if platform.system().lower() == "windows" and installer == "pip-sync":
        pytest.skip("pip-sync + CLI runner not supported on Windows")
    environment = pip_compile.default_environment
    constraint_env = pip_compile.test_environment
    # Start with a clean slate, no environments
    assert not pip_compile.virtualenv_exists(environment=environment)
    assert not pip_compile.virtualenv_exists(environment=constraint_env)
    pip_compile.update_environment_resolver(environment=environment, resolver=resolver)
    pip_compile.update_environment_installer(environment=environment, installer=installer)
    # Invoke the constraint environment first. The default environment should not exist yet
    result = pip_compile.invoke_environment(environment=constraint_env)
    if result.exit_code != 0:
        raise result.exception
    assert not pip_compile.virtualenv_exists(environment=environment)
    assert pip_compile.is_installed(environment=constraint_env, package="pytest")
    assert "requests" not in environment.piptools_lock_file.read_text()
    assert not pip_compile.is_installed(environment=constraint_env, package="requests")
    # Now add the `requests` package to the default environment, and invoke the constraint env
    # The `requests` package should be installed in both environments and both lockfiles updated
    pip_compile.toml_doc["project"]["dependencies"] += ["requests"]
    pip_compile.update_pyproject()
    pip_compile.invoke_environment(environment=constraint_env)
    assert "requests" in environment.piptools_lock_file.read_text()
    assert "requests" in constraint_env.piptools_lock_file.read_text()
    assert pip_compile.is_installed(environment=environment, package="requests")
    assert pip_compile.is_installed(environment=constraint_env, package="requests")



"""
Base classes for hatch-pip-compile
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from hatchling.dep.core import dependencies_in_sync
from packaging.requirements import Requirement

if TYPE_CHECKING:
    from hatch_pip_compile.plugin import PipCompileEnvironment


class HatchPipCompileBase:
    """
    Base Class for hatch-pip-compile tools
    """

    pypi_dependencies: ClassVar[list[str]] = []

    def __init__(self, environment: PipCompileEnvironment) -> None:
        """
        Inject the environment into the base class
        """
        self.environment = environment
        self.pypi_dependencies_installed = False

    def install_pypi_dependencies(self) -> None:
        """
        Install the resolver from PyPI
        """
        if not self.pypi_dependencies:
            return
        elif self.pypi_dependencies_installed:
            return
        with self.environment.safe_activation():
            in_sync = dependencies_in_sync(
                requirements=[Requirement(item) for item in self.pypi_dependencies],
                sys_path=self.environment.virtual_env.sys_path,
                environment=self.environment.virtual_env.environment,
            )
            if not in_sync:
                self.environment.plugin_check_command(
                    self.environment.construct_pip_install_command(self.pypi_dependencies)
                )
            self.pypi_dependencies_installed = True



"""
hatch-pip-compile
"""

from hatch_pip_compile.__about__ import __application__, __author__, __email__, __version__

__all__ = [
    "__application__",
    "__version__",
    "__author__",
    "__email__",
]



"""
Testing the hatch-pip-compile CLI
"""

from unittest.mock import Mock

import click
import pytest
from click.testing import CliRunner

from hatch_pip_compile import __version__
from hatch_pip_compile.cli import HatchCommandRunner, cli
from tests.conftest import PipCompileFixture


def test_cli_help() -> None:
    """
    Test the CLI with the --help flag
    """
    runner = CliRunner()
    result = runner.invoke(cli=cli, args=["--help"])
    assert result.exit_code == 0


def test_cli_version() -> None:
    """
    Test the CLI with the --version flag
    """
    runner = CliRunner()
    result = runner.invoke(cli=cli, args=["--version"])
    assert result.exit_code == 0
    assert f"hatch-pip-compile, version {__version__}" in result.output


def test_cli_no_args_mocked(pip_compile: PipCompileFixture, subprocess_run: Mock) -> None:
    """
    Test the CLI with no arguments - mock the result
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        _ = runner.invoke(cli=cli)
        assert subprocess_run.call_count == 1
        subprocess_run.assert_called_once()
        subprocess_run.assert_called_with(
            args=["hatch", "env", "show", "--json"], capture_output=True, check=True
        )


def test_cli_no_args(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with no arguments
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli)
        assert result.exit_code == 0
        assert "hatch-pip-compile: Targeting environments: default" in result.output
        assert (
            "hatch-pip-compile: Running `hatch env run --env default -- python --version`"
            in result.output
        )


def test_cli_bad_env(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with a non-existent environment
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli, args=["bad_env"])
        assert result.exit_code != 0
        assert "error" in result.output.lower()
        assert (
            "The following environments are not supported or unknown: bad_env. "
            "Supported environments are: default, docs, lint, misc, test"
        ) in result.output


def test_cli_test_env(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with the `test` argument
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli, args=["test"])
        assert result.exit_code == 0
        assert "hatch-pip-compile: Targeting environments: test" in result.output
        assert (
            "hatch-pip-compile: Running `hatch env run --env test -- python --version`"
            in result.output
        )


def test_cli_all(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with the `--all` argument
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli, args=["--all"])
        assert result.exit_code == 0
        assert (
            "hatch-pip-compile: Targeting environments: default, docs, lint, misc, test"
            in result.output
        )


def test_cli_upgrade(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with the `--upgrade` argument
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli, args=["--upgrade"])
        assert result.exit_code == 0
        assert "hatch-pip-compile: Upgrading all dependencies" in result.output


def test_cli_upgrade_packages(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with the `--upgrade-package` argument
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(
            cli=cli, args=["--upgrade-package", "requests", "--upgrade-package", "urllib3"]
        )
        assert result.exit_code == 0
        assert "hatch-pip-compile: Upgrading packages: requests, urllib3" in result.output


def test_cli_upgrade_test(pip_compile: PipCompileFixture) -> None:
    """
    Test the full CLI with the `--upgrade` argument for the `test` environment
    """
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=pip_compile.isolation):
        result = runner.invoke(cli=cli, args=["test", "--upgrade"])
        assert result.exit_code == 0
        assert "hatch-pip-compile: Upgrading all dependencies" in result.output
        assert "hatch-pip-compile: Targeting environments: test" in result.output


def test_command_runner_supported_environments(
    monkeypatch: pytest.MonkeyPatch,
    pip_compile: PipCompileFixture,
) -> None:
    """
    Test the `supported_environments` attribute
    """
    with pip_compile.chdir():
        command_runner = HatchCommandRunner(
            environments=["test"],
            upgrade=True,
            upgrade_packages=[],
        )
        assert command_runner.supported_environments == {"default", "test", "lint", "docs", "misc"}


def test_command_runner_non_supported_environments(
    monkeypatch: pytest.MonkeyPatch,
    pip_compile: PipCompileFixture,
) -> None:
    """
    Test that a bad environment raises a `BadParameter` exception
    """
    with pip_compile.chdir():
        with pytest.raises(
            click.BadParameter,
            match=(
                "The following environments are not supported or unknown: bad_env. "
                "Supported environments are: default, docs, lint, misc, test"
            ),
        ):
            _ = HatchCommandRunner(
                environments=["bad_env"],
                upgrade=True,
                upgrade_packages=[],
            )



"""
Testing the `lock` module
"""
from textwrap import dedent

from packaging.requirements import Requirement
from packaging.version import Version

from tests.conftest import PipCompileFixture


def test_lockfile_python_version(
    pip_compile: PipCompileFixture,
):
    """
    Test the expected Python version from the lockfiles
    """
    assert pip_compile.default_environment.piptools_lock.lock_file_version == Version("3.11")
    assert pip_compile.test_environment.piptools_lock.lock_file_version == Version("3.11")


def test_lockfile_dependencies(
    pip_compile: PipCompileFixture,
):
    """
    Test the expected dependencies from reading the lockfiles
    """
    assert set(pip_compile.default_environment.piptools_lock.read_header_requirements()) == {
        Requirement("hatch")
    }
    assert set(pip_compile.test_environment.piptools_lock.read_header_requirements()) == {
        Requirement("pytest"),
        Requirement("pytest-cov"),
        Requirement("hatch"),
    }


def test_compare_requirements_match(pip_compile: PipCompileFixture):
    """
    Test the `compare_requirements` method with a match
    """
    default_check = pip_compile.default_environment.piptools_lock.compare_requirements(
        requirements=[Requirement("hatch")],
    )
    assert default_check is True


def test_compare_requirements_mismatch(pip_compile: PipCompileFixture):
    """
    Test the `compare_requirements` method with a mismatch
    """
    test_check = pip_compile.test_environment.piptools_lock.compare_requirements(
        requirements=[Requirement("hatch")],
    )
    assert test_check is False


def test_read_lock_requirements(pip_compile: PipCompileFixture) -> None:
    """
    Test the `read_lock_requirements` method
    """
    linting_env = pip_compile.reload_environment("lint")
    requirements = linting_env.piptools_lock.read_lock_requirements()
    string_requirements = set(map(str, requirements))
    assert string_requirements == {
        "mypy==1.7.1",
        "mypy-extensions==1.0.0",
        "ruff==0.1.6",
        "typing-extensions==4.8.0",
    }


def test_replace_temporary_lockfile_windows(pip_compile: PipCompileFixture) -> None:
    """
    Regex Replace Temporary File Path: Windows
    """
    lock_raw = r"""
    httpx==0.22.0
        # via -r C:\Users\xxx\AppData\Local\Temp\tmp_kn984om\default.in
    """
    lock_body = dedent(lock_raw).strip()
    cleaned_text = pip_compile.default_environment.piptools_lock.replace_temporary_lockfile(
        lock_body
    )
    expected_raw = r"""
    httpx==0.22.0
        # via hatch.envs.default
    """
    assert cleaned_text == dedent(expected_raw).strip()


def test_replace_temporary_lockfile_unix(pip_compile: PipCompileFixture) -> None:
    """
    Regex Replace Temporary File Path: Unix
    """
    lock_raw = r"""
    httpx==0.22.0
        # via -r /tmp/tmp_kn984om/default.in
    """
    lock_body = dedent(lock_raw).strip()
    cleaned_text = pip_compile.default_environment.piptools_lock.replace_temporary_lockfile(
        lock_body
    )
    expected_raw = r"""
    httpx==0.22.0
        # via hatch.envs.default
    """
    assert cleaned_text == dedent(expected_raw).strip()



"""
Package + Dependency Installers
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import ClassVar

from hatch.env.utils import add_verbosity_flag

from hatch_pip_compile.base import HatchPipCompileBase


class PluginInstaller(HatchPipCompileBase, ABC):
    """
    Package Installer for the plugin

    This abstract base class is used to define the interface for
    how the plugin should install packages and dependencies.
    """

    @abstractmethod
    def install_dependencies(self) -> None:
        """
        Install the dependencies
        """

    def sync_dependencies(self) -> None:
        """
        Sync the dependencies - same as `install_dependencies`
        """
        self.install_pypi_dependencies()
        self.install_dependencies()

    def construct_pip_install_command(self, args: list[str]) -> list[str]:
        """
        Construct a `pip install` command with the given arguments
        """
        return self.environment.construct_pip_install_command(args)

    def install_project(self) -> None:
        """
        Install the project (`--no-deps`)
        """
        self.install_pypi_dependencies()
        with self.environment.safe_activation():
            self.environment.plugin_check_command(
                self.construct_pip_install_command(args=["--no-deps", str(self.environment.root)])
            )

    def install_project_dev_mode(self) -> None:
        """
        Install the project in editable mode (`--no-deps`)
        """
        self.install_pypi_dependencies()
        with self.environment.safe_activation():
            self.environment.plugin_check_command(
                self.construct_pip_install_command(
                    args=["--no-deps", "--editable", str(self.environment.root)]
                )
            )


class PipInstaller(PluginInstaller):
    """
    Plugin Installer for `pip`
    """

    def install_dependencies(self) -> None:
        """
        Install the dependencies with `pip`
        """
        self.install_pypi_dependencies()
        with self.environment.safe_activation():
            if not self.environment.piptools_lock_file.exists():
                return
            extra_args = self.environment.config.get("pip-compile-install-args", [])
            args = [*extra_args, "--requirement", str(self.environment.piptools_lock_file)]
            install_command = self.construct_pip_install_command(args=args)
            self.environment.plugin_check_command(install_command)


class UvInstaller(PipInstaller):
    """
    Plugin Installer for `uv`
    """

    pypi_dependencies: ClassVar[list[str]] = ["uv"]

    def construct_pip_install_command(self, args: list[str]) -> list[str]:
        """
        Construct a `pip install` command with the given arguments
        """
        command = [
            "python",
            "-m",
            "uv",
            "pip",
            "install",
        ]
        add_verbosity_flag(command, self.environment.verbosity, adjustment=-1)
        command.extend(args)
        return command


class PipSyncInstaller(PluginInstaller):
    """
    Plugin Installer for `pip-sync`
    """

    pypi_dependencies: ClassVar[list[str]] = ["pip-tools"]

    def install_dependencies(self) -> None:
        """
        Install the dependencies with `pip-sync`

        In the event that there are no dependencies, pip-sync will
        uninstall everything in the environment before deleting the
        lockfile.
        """
        self.install_pypi_dependencies()
        cmd = [
            self.environment.virtual_env.python_info.executable,
            "-m",
            "piptools",
            "sync",
            "--verbose"
            if self.environment.config.get("pip-compile-verbose", None) is True
            else "--quiet",
            "--python-executable",
            str(self.environment.virtual_env.python_info.executable),
        ]
        if not self.environment.dependencies:
            self.environment.piptools_lock_file.write_text("")
        extra_args = self.environment.config.get("pip-compile-install-args", [])
        cmd.extend(extra_args)
        cmd.append(str(self.environment.piptools_lock_file))
        self.environment.plugin_check_command(cmd)
        if not self.environment.dependencies:
            self.environment.piptools_lock_file.unlink()

    def _full_install(self) -> None:
        """
        Run the full install process

        1) Run pip-compile (if necessary)
        2) Run pip-sync
        3) (re)install project
        """
        with self.environment.safe_activation():
            self.environment.run_pip_compile()
            self.install_dependencies()
        if not self.environment.skip_install:
            if self.environment.dev_mode:
                super().install_project_dev_mode()
            else:
                super().install_project()

    def sync_dependencies(self):
        """
        Sync dependencies
        """
        self._full_install()

    def install_project(self):
        """
        Install the project the first time

        The same implementation as `_full_install`
        due to the way `pip-sync` uninstalls our root package
        """
        self._full_install()

    def install_project_dev_mode(self):
        """
        Install the project the first time in dev mode

        The same implementation as `_full_install`
        due to the way `pip-sync` uninstalls our root package
        """
        self._full_install()



"""
hatch-pip-compile plugin
"""

import functools
import hashlib
import logging
import os
import pathlib
import shutil
import tempfile
from subprocess import CompletedProcess
from typing import Any, ClassVar, Dict, List, Optional, Type, Union

import hatch.cli
from hatch.env.virtual import VirtualEnvironment
from hatch.utils.platform import Platform
from hatchling.dep.core import dependencies_in_sync

from hatch_pip_compile.exceptions import HatchPipCompileError
from hatch_pip_compile.installer import PipInstaller, PipSyncInstaller, PluginInstaller, UvInstaller
from hatch_pip_compile.lock import PipCompileLock
from hatch_pip_compile.resolver import BaseResolver, PipCompileResolver, UvResolver

logger = logging.getLogger(__name__)


class PipCompileEnvironment(VirtualEnvironment):
    """
    Virtual Environment supported by pip-compile
    """

    PLUGIN_NAME: ClassVar[str] = "pip-compile"
    default_env_name: ClassVar[str] = "default"
    dependency_resolvers: ClassVar[Dict[str, Type[BaseResolver]]] = {
        "pip-compile": PipCompileResolver,
        "uv": UvResolver,
    }
    dependency_installers: ClassVar[Dict[str, Type[PluginInstaller]]] = {
        "pip": PipInstaller,
        "pip-sync": PipSyncInstaller,
        "uv": UvInstaller,
    }

    def __repr__(self):
        """
        Get representation of PipCompileEnvironment
        """
        return f"<{self.__class__.__name__} - {self.name}>"

    def __init__(self, *args, **kwargs) -> None:
        """
        Initialize PipCompileEnvironment with extra attributes
        """
        super().__init__(*args, **kwargs)
        lock_filename_config = self.config.get("lock-filename")
        if lock_filename_config is None:
            if self.name == self.default_env_name:
                lock_filename = "requirements.txt"
            else:
                lock_filename = f"requirements/requirements-{self.name}.txt"
        else:
            with self.metadata.context.apply_context(self.context):
                lock_filename = self.metadata.context.format(lock_filename_config)
        self.piptools_lock_file = self.root / lock_filename
        self.piptools_lock = PipCompileLock(environment=self)
        install_method = self.config.get("pip-compile-installer", "pip")
        resolve_method = self.config.get("pip-compile-resolver", "pip-compile")
        if install_method not in self.dependency_installers.keys():
            msg = (
                f"Invalid pip-compile-installer: {install_method} - "
                f"must be one of {', '.join(self.dependency_installers.keys())}"
            )
            raise HatchPipCompileError(msg)
        if resolve_method not in self.dependency_resolvers.keys():
            msg = (
                f"Invalid pip-compile-resolver: {resolve_method} - "
                f"must be one of {', '.join(self.dependency_resolvers.keys())}"
            )
            raise HatchPipCompileError(msg)
        resolver_class = self.dependency_resolvers[resolve_method]
        installer_class = self.dependency_installers[install_method]
        self.resolver: BaseResolver = resolver_class(environment=self)
        self.installer: PluginInstaller = installer_class(environment=self)

    @staticmethod
    def get_option_types() -> Dict[str, Any]:
        """
        Get option types
        """
        return {  # pragma: no cover
            "lock-filename": str,
            "pip-compile-hashes": bool,
            "pip-compile-args": List[str],
            "pip-compile-constraint": str,
            "pip-compile-installer": str,
            "pip-compile-install-args": List[str],
            "pip-compile-resolver": str,
        }

    def dependency_hash(self) -> str:
        """
        Get the dependency hash
        """
        self.run_pip_compile()
        hatch_hash = super().dependency_hash()
        if not self.dependencies:
            return hatch_hash
        else:
            lockfile_hash = self.piptools_lock.get_file_content_hash()
            return hashlib.sha256(f"{hatch_hash}-{lockfile_hash}".encode()).hexdigest()

    def run_pip_compile(self) -> None:
        """
        Run pip-compile if necessary
        """
        self.prepare_environment()
        if not self.lockfile_up_to_date:
            with self.safe_activation():
                self.resolver.install_pypi_dependencies()
                if self.piptools_lock_file.exists():
                    _ = self.piptools_lock.compare_python_versions(
                        verbose=self.config.get("pip-compile-verbose", None)
                    )
                self.pip_compile_cli()

    def pip_compile_cli(self) -> None:
        """
        Run pip-compile
        """
        if not self.dependencies:
            self.piptools_lock_file.unlink(missing_ok=True)
            self.lockfile_up_to_date = True
            return
        no_compile = bool(os.getenv("PIP_COMPILE_DISABLE"))
        if no_compile:
            msg = "hatch-pip-compile is disabled but attempted to run a lockfile update."
            raise HatchPipCompileError(msg)
        with tempfile.TemporaryDirectory() as tmpdir:
            tmp_path = pathlib.Path(tmpdir)
            input_file = tmp_path / f"{self.name}.in"
            output_file = tmp_path / "lock.txt"
            input_file.write_text("\n".join([*self.dependencies, ""]))
            if self.piptools_lock_file.exists():
                shutil.copy(self.piptools_lock_file, output_file)
            self.piptools_lock_file.parent.mkdir(exist_ok=True, parents=True)
            cmd = self.resolver.get_pip_compile_args(
                input_file=input_file,
                output_file=output_file,
            )
            self.plugin_check_command(cmd)
            self.piptools_lock.process_lock(lockfile=output_file)
            shutil.move(output_file, self.piptools_lock_file)
        self.lockfile_up_to_date = True

    def install_project(self) -> None:
        """
        Install the project (`--no-deps`)
        """
        self.installer.install_project()

    def install_project_dev_mode(self) -> None:
        """
        Install the project in editable mode (`--no-deps`)
        """
        self.installer.install_project_dev_mode()

    @functools.cached_property
    def lockfile_up_to_date(self) -> bool:
        """
        Check if the lockfile is up-to-date

        Behavior
        --------
        1) If there are no dependencies and no lock file, exit early and return True.
        2) If the constraint file / environment is out of date, sync it and return False.
        3) If there are no dependencies and a lock file, return False.
        4) If there are dependencies and no lock file, return False.
        5) If a force upgrade is requested, return False.
        6) If there are dependencies and a lock file...
            a) If there is a constraint file...
                i) If the file is valid but the SHA is different, return False.
            b) If the lock file dependencies aren't current, return False.
            c) If the lock file dependencies are current but the lockfile
               has a different sha than its constraints file, return False.
        7) Otherwise, return True.
        """
        upgrade = os.getenv("PIP_COMPILE_UPGRADE") or False
        upgrade_packages = os.getenv("PIP_COMPILE_UPGRADE_PACKAGE") or False
        pip_compile_force = bool(os.getenv("__PIP_COMPILE_FORCE__"))
        force_upgrade = any(
            [
                upgrade is not False,
                upgrade_packages is not False,
                pip_compile_force is not False,
            ]
        )
        if not self.dependencies and not self.piptools_lock_file.exists():
            return True
        if self.piptools_constraints_file:
            valid_constraint = self.validate_constraints_file(
                constraints_file=self.piptools_constraints_file, environment=self.constraint_env
            )
            if not valid_constraint:
                return False
        if not self.dependencies and self.piptools_lock_file.exists():
            return False
        elif force_upgrade:
            return False
        elif self.dependencies and not self.piptools_lock_file.exists():
            return False
        elif self.dependencies and self.piptools_lock_file.exists():
            if self.piptools_constraints_file:
                current_sha = self.constraint_env.piptools_lock.get_file_content_hash()
                sha_match = self.piptools_lock.compare_constraint_sha(sha=current_sha)
                if sha_match is False:
                    return False
            expected_dependencies = self.piptools_lock.compare_requirements(
                requirements=self.dependencies_complex
            )
            if not expected_dependencies:
                return False
        return True

    def dependencies_in_sync(self):
        """
        Whether the dependencies are in sync
        """
        if not self.lockfile_up_to_date:
            return False
        else:
            with self.safe_activation():
                return dependencies_in_sync(
                    self.piptools_lock.read_lock_requirements(),
                    sys_path=self.virtual_env.sys_path,
                    environment=self.virtual_env.environment,
                )

    def sync_dependencies(self) -> None:
        """
        Sync dependencies
        """
        self.run_pip_compile()
        self.installer.sync_dependencies()

    @property
    def piptools_constraints_file(self) -> Optional[pathlib.Path]:
        """
        Get the constraint file path
        """
        if self.constraint_env.name == self.name:
            return None
        else:
            return self.constraint_env.piptools_lock_file

    def get_piptools_environment(self, environment_name: str) -> "PipCompileEnvironment":
        """
        Get a `PipCompileEnvironment` instance for an environment
        other than the current instance. This is useful
        for recursively checking other environments for lock file
        validity and defining inheritance.
        """
        if environment_name not in self.pipools_environment_dict.keys():
            error_message = (
                f"[hatch-pip-compile] The environment {environment_name} does not exist."
            )
            raise HatchPipCompileError(error_message)
        if isinstance(self.app, hatch.cli.Application):
            env = self.app.get_environment(env_name=environment_name)
        else:
            env = PipCompileEnvironment(
                root=self.root,
                metadata=self.metadata,
                name=environment_name,
                config=self.pipools_environment_dict.get(environment_name, {}),
                matrix_variables=self.matrix_variables,
                data_directory=self.data_directory,
                isolated_data_directory=self.isolated_data_directory,
                platform=Platform(),
                verbosity=self.verbosity,
                app=self.app,
            )
        return env

    @functools.cached_property
    def constraint_env(self) -> "PipCompileEnvironment":
        """
        Get the constraint environment
        """
        constraint_env = self.config.get("pip-compile-constraint")
        if not constraint_env:
            return self
        elif self.name == constraint_env:
            return self
        environment = self.get_piptools_environment(environment_name=constraint_env)
        if environment.config.get("type") != self.PLUGIN_NAME:
            logger.error("The constraint environment is not a hatch-pip-compile environment.")
            return self
        elif not environment.dependencies:
            return self
        else:
            return environment

    def validate_constraints_file(
        self, constraints_file: pathlib.Path, environment: "PipCompileEnvironment"
    ) -> bool:
        """
        Validate the constraints file

        Parameters
        ----------
        constraints_file : pathlib.Path
            The lock file
        environment : PipCompileEnvironment
            The environment to validate against

        Returns
        -------
        bool
            Whether the constraints file is valid
        """
        if not constraints_file.exists():
            self.constraint_env.run_pip_compile()
            return False
        else:
            up_to_date = environment.piptools_lock.compare_requirements(
                requirements=environment.dependencies_complex
            )
            if not up_to_date:
                self.constraint_env.run_pip_compile()
                return False
        return True

    @property
    def pipools_environment_dict(self) -> Dict[str, Any]:
        """
        Get the environment dictionary
        """
        return self.metadata.hatch.config.get("envs", {})

    def plugin_check_command(
        self, command: Union[str, List[str]], *, shell: bool = False, **kwargs: Any
    ) -> CompletedProcess:
        """
        Run a command from the virtualenv
        """
        with self.safe_activation():
            return self.virtual_env.platform.check_command(
                command=command,
                shell=shell,
                **kwargs,
            )

    def virtualenv_exists(self) -> bool:
        """
        Check if the virtualenv exists
        """
        try:
            _ = self.virtual_env.executables_directory
            return True
        except OSError:
            return False

    def prepare_environment(self) -> None:
        """
        Prepare the environment

        Ideally, we could access the `self.app` attribute,
        but this ultimately leads to a recursion error when
        `self.app.prepare_environment()` is called.
        """
        if not self.virtualenv_exists():
            self.create()
            if not self.dependencies_in_sync():
                self.sync_dependencies()
            if not self.skip_install:
                if self.dev_mode:
                    self.install_project_dev_mode()
                else:
                    self.install_project()
